package at.jku.students.sort;

import java.util.ArrayList;
import java.util.Comparator;

import at.jku.ssw.sort.Sorter;
import at.jku.ssw.sort.Sorter.SortingOrder;
import at.jku.ssw.util.Out;

@SuppressWarnings({ "rawtypes", "unchecked" })
public class SortSearchMain {

	public static void main(String[] args) {
		boolean b = true;

		// test empty
		Out.println("Testing empty array sort with heap sort....");
		try {
			testEmpty(new HeapSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		Out.println("Testing empty array sort with merge sort....");
		try {
			testEmpty(new MergeSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		// test equal element array
		Out.println("Testing array with equal elements sort with heap sort....");
		try {
			testEqual(new HeapSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		Out.println("Testing array with equal elements sort with merhr sort....");
		try {
			testEqual(new MergeSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		// test wrongly sorted array
		Out.println("Testing array with reversed elements sort with heap sort....");
		try {
			testWronglySorted(new HeapSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		Out.println("Testing array with reversed elements sort with merge sort....");
		try {
			testWronglySorted(new MergeSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		// test arbitrary sorted array
		Out.println("Testing array with arbitrary elements sort with heap sort....");
		try {
			arbitrarySorted(new HeapSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		Out.println("Testing array with arbitrary elements sort with merge sort....");
		try {
			arbitrarySorted(new MergeSorter());
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		Out.println("Finally testing searching with binary search....");
		try {
			testSearch();
			Out.println("Tests succeeded.");
		} catch (Throwable t) {
			b = false;
			Out.println("\t[Error] " + t.getMessage());
		}

		if (b) {
			Out.println("\n\nEverything seems to be fine. Congrats! ;-)");
		} else {
			Out.println("\n\nThere have been failing test cases.......FIX THEM PLEASE ;-)");
		}
	}

	private static void testSearch() {
		BinarySearcher bs = new BinarySearcher();

		// test assignment
		Out.println("Test search assignment example....");
		Object[] o1 = createAndSort(new MergeSorter(), INTEGER_COMPARATOR, SortingOrder.ASCENDING, 1, 3, 4, 6, 7, 8, 10, 13, 14);
		assertEquals(indexOf(o1, INTEGER_COMPARATOR, 4), bs.search(o1, 4, INTEGER_COMPARATOR));

		// test arbitrary sorting included
		Out.println("Test search arbitrary included....");
		o1 = createAndSort(new MergeSorter(), INTEGER_COMPARATOR, SortingOrder.ASCENDING, 42, 11, 1, 0, -1, 100, 9999, 0XDEADBEEF,
				1234, 5555522, 7673230, 99944121, 23332478, 4423);
		assertEquals(indexOf(o1, INTEGER_COMPARATOR, 42), bs.search(o1, 42, INTEGER_COMPARATOR));

		// test empty
		Out.println("Test search empty source....");
		o1 = createAndSort(new MergeSorter(), INTEGER_COMPARATOR, SortingOrder.ASCENDING);
		assertEquals(indexOf(o1, INTEGER_COMPARATOR, 42), bs.search(o1, 42, INTEGER_COMPARATOR));
	}

	private static int indexOf(Object[] items, Comparator c, Object key) {
		ArrayList<Object> a = new ArrayList<>();
		for (Object o : items) {
			a.add(o);
		}
		a.sort(c);
		return a.indexOf(key);
	}

	private static final Comparator EMPTY_COMPARATOR = new Comparator() {

		@Override
		public int compare(Object o1, Object o2) {
			return 0;
		}
	};
	private static final Comparator INTEGER_COMPARATOR = new Comparator() {

		@Override
		public int compare(Object o1, Object o2) {
			return Integer.compare((Integer) o1, (Integer) o2);
		}
	};
	private static void testEmpty(Sorter s) {
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING), EMPTY_COMPARATOR, SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING), EMPTY_COMPARATOR, SortingOrder.DECENDING);
	}

	private static void testEqual(Sorter s) {
		Integer i = new Integer(0);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING, i, i, i, i, i, i, i, i, i, i, i, i, i, i),
				INTEGER_COMPARATOR, SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING, i, i, i, i, i, i, i, i, i, i, i, i, i, i),
				INTEGER_COMPARATOR, SortingOrder.DECENDING);
	}

	private static void testWronglySorted(Sorter s) {
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10), INTEGER_COMPARATOR,
				SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1), INTEGER_COMPARATOR,
				SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10),
				INTEGER_COMPARATOR, SortingOrder.DECENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING, 10, 9, 8, 7, 6, 5, 4, 3, 2, 1),
				INTEGER_COMPARATOR, SortingOrder.DECENDING);
	}

	private static void arbitrarySorted(Sorter s) {
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING, 42, 11, 1, 0, -1, 100, 9999, 0XDEADBEEF, 1234,
				5555522, 7673230, 99944121, 23332478, 4423), INTEGER_COMPARATOR, SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.ASCENDING, 42, 11, 1, 0, -1, 100, 9999, 0XDEADBEEF, 1234,
				5555522, 7673230, 99944121, 23332478, 4423), INTEGER_COMPARATOR, SortingOrder.ASCENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING, 42, 11, 1, 0, -1, 100, 9999, 0XDEADBEEF, 1234,
				5555522, 7673230, 99944121, 23332478, 4423), INTEGER_COMPARATOR, SortingOrder.DECENDING);
		assertSorted(createAndSort(s, INTEGER_COMPARATOR, SortingOrder.DECENDING, 42, 11, 1, 0, -1, 100, 9999, 0XDEADBEEF, 1234,
				5555522, 7673230, 99944121, 23332478, 4423), INTEGER_COMPARATOR, SortingOrder.DECENDING);
	}

	private static Object[] createAndSort(Sorter s, Comparator c, SortingOrder ordering, Object... os) {

		Object oo[] = new Object[os.length];
		for (int i = 0; i < oo.length; i++) {
			oo[i] = os[i];
		}
		Out.println("\t[Input before sorting] array=" + createString(oo));
		oo = s.sort(oo, c, ordering);
		Out.println("\t[Input after  sorting] array=" + createString(oo) + " order is = " + ordering);
		return oo;
	}

	private static void test(boolean b, String failed) {
		if (!b) {
			throw new RuntimeException(failed);
		}
	}

	private static void assertEquals(Object a, Object b) {
		test(a.equals(b), "Objects a " + a + " b " + b + " are not equal.");
	}

	private static void assertSorted(Object[] items, Comparator c, SortingOrder ordering) {
		ArrayList<Object> real = new ArrayList<>();
		for (Object o : items) {
			real.add(o);
		}
		real.sort(c);
		if (ordering == SortingOrder.DECENDING) {
			for (int i = items.length - 1, j = 0; i >= 0; i--, j++) {
				assertEquals(real.get(i), items[j]);
			}
		} else {
			for (int i = 0; i < items.length; i++) {
				assertEquals(real.get(i), items[i]);
			}
		}
	}

	private static String createString(Object[] items) {
		StringBuilder sb = new StringBuilder();
		sb.append("[");
		for (int i = 0; i < items.length; i++) {
			sb.append(items[i]);
			if (i < items.length - 1) {
				sb.append(";");
			}
		}
		sb.append("]");
		return sb.toString();
	}
}
